# make / work consistently between python 2.x and 3.x
# https://www.python.org/dev/peps/pep-0238/
from __future__ import division
from functools import reduce
import math


def safeSum(values):
    safeValues = [v for v in values if v is not None]
    if safeValues:
        return sum(safeValues)


def safeDiff(values):
    safeValues = [v for v in values if v is not None]
    if safeValues:
        values = list(map(lambda x: x*-1, safeValues[1:]))
        values.insert(0, safeValues[0])
        return sum(values)


def safeLen(values):
    return len([v for v in values if v is not None])


def safeDiv(a, b):
    if a is None: return None
    if b in (0,None): return None
    return a / b


def safeExp(a):
    try:
        return math.exp(a)
    except TypeError:
        return None


def safePow(a, b):
    if a is None or b is None:
        return None
    try:
        result = math.pow(a, b)
    except (ValueError, OverflowError):
        return None
    return result


def safeMul(*factors):
    if None in factors:
        return None

    factors = [float(x) for x in factors]
    product = reduce(lambda x,y: x*y, factors)
    return product


def safeSubtract(a,b):
    if a is None or b is None:
        return None
    return float(a) - float(b)


def safeAvg(values):
    safeValues = [v for v in values if v is not None]
    if safeValues:
        return sum(safeValues) / len(safeValues)


def safeAvgZero(values):
    if values:
        return sum([0 if v is None else v for v in values]) / len(values)


def safeMedian(values):
    safeValues = [v for v in values if v is not None]
    if safeValues:
        sortedVals = sorted(safeValues)
        mid = len(sortedVals) // 2
        if len(sortedVals) % 2 == 0:
            return float(sortedVals[mid-1] + sortedVals[mid]) / 2
        else:
            return sortedVals[mid]


def safeStdDev(a):
    sm = safeSum(a)
    ln = safeLen(a)
    avg = safeDiv(sm,ln)
    if avg is None:
        return None
    sum = 0
    safeValues = [v for v in a if v is not None]
    for val in safeValues:
        sum = sum + (val - avg) * (val - avg)
    return math.sqrt(sum/ln)


def safeLast(values):
    for v in reversed(values):
        if v is not None: return v


def safeMin(values, default=None):
    safeValues = [v for v in values if v is not None]
    if safeValues:
        return min(safeValues)
    else:
        return default


def safeMax(values, default=None):
    safeValues = [v for v in values if v is not None]
    if safeValues:
        return max(safeValues)
    else:
        return default


def safeMap(function, values):
    safeValues = [v for v in values if v is not None]
    if safeValues:
        return [function(x) for x in safeValues]


def safeAbs(value):
    if value is None:
        return None
    return abs(value)



from django.contrib import admin
from graphite.account.models import Profile,MyGraph


class MyGraphAdmin(admin.ModelAdmin):
    list_display = ('profile','name')
    list_filter = ('profile',)


admin.site.register(Profile)
admin.site.register(MyGraph, MyGraphAdmin)



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import pytz
from datetime import datetime, timedelta, datetime as datetimetype
from django.conf import settings

months = ['jan','feb','mar','apr','may','jun','jul','aug','sep','oct','nov','dec']
weekdays = ['sun','mon','tue','wed','thu','fri','sat']
SECONDS_STRING = 'seconds'
MINUTES_STRING = 'minutes'
HOURS_STRING = 'hours'
DAYS_STRING = 'days'
WEEKS_STRING = 'weeks'
MONTHS_STRING = 'months'
YEARS_STRING = 'years'


def parseATTime(s, tzinfo=None, now=None):
    if tzinfo is None:
        tzinfo = pytz.timezone(settings.TIME_ZONE)
    if isinstance(s, datetimetype):
        if s.tzinfo:
            return s.astimezone(tzinfo)
        return tzinfo.localize(s)

    s = s.strip().lower().replace('_','').replace(',','').replace(' ','')
    if s.isdigit():
        if len(s) == 8 and int(s[:4]) > 1900 and int(s[4:6]) < 13 and int(s[6:]) < 32:
            pass  # Fall back because its not a timestamp, its YYYYMMDD form
        else:
            return datetime.fromtimestamp(int(s),tzinfo)
    if '+' in s:
        ref,offset = s.split('+',1)
        offset = '+' + offset
    elif '-' in s:
        ref,offset = s.split('-',1)
        offset = '-' + offset
    else:
        ref,offset = s,''

    return tzinfo.normalize(parseTimeReference(ref, tzinfo, now) + parseTimeOffset(offset))


def parseTimeReference(ref, tzinfo=None, now=None):
    if tzinfo is None:
        tzinfo = pytz.timezone(settings.TIME_ZONE)
    if isinstance(ref, datetimetype):
        if ref.tzinfo:
            return ref.astimezone(tzinfo)
        return tzinfo.localize(ref)

    if now is None:
        now = datetime.now(tzinfo)
    else:
        now = parseATTime(now, tzinfo)

    if not ref or ref == 'now':
        return now

    rawRef = ref

    # Time-of-day reference
    i = ref.find(':')
    hour,minute = 0,0
    if 0 < i < 3:
        hour = int( ref[:i] )
        minute = int( ref[i+1:i+3] )
        ref = ref[i+3:]
        if ref[:2] == 'am':
            ref = ref[2:]
        elif ref[:2] == 'pm':
            hour = (hour + 12) % 24
            ref = ref[2:]

    # Xam or XXam
    i = ref.find('am')
    if 0 < i < 3:
        hour = int( ref[:i] )
        ref = ref[i+2:]

    # Xpm or XXpm
    i = ref.find('pm')
    if 0 < i < 3:
        hour = (int( ref[:i] ) + 12) % 24
        ref = ref[i+2:]

    if ref.startswith('noon'):
        hour,minute = 12,0
        ref = ref[4:]
    elif ref.startswith('midnight'):
        hour,minute = 0,0
        ref = ref[8:]
    elif ref.startswith('teatime'):
        hour,minute = 16,0
        ref = ref[7:]

    refDate = now.replace(hour=hour,minute=minute,second=0,microsecond=0,tzinfo=None)

    # Day reference
    if ref in ('yesterday','today','tomorrow'):  # yesterday, today, tomorrow
        if ref == 'yesterday':
            refDate -= timedelta(days=1)
        elif ref == 'tomorrow':
            refDate += timedelta(days=1)

    elif ref.count('/') == 2:  # MM/DD/YY[YY]
        m,d,y = map(int,ref.split('/'))
        if y < 1900:
            y += 1900
        if y < 1970:
            y += 100
        refDate = datetime(year=y,month=m,day=d,hour=hour,minute=minute)

    elif len(ref) == 8 and ref.isdigit():  # YYYYMMDD
        refDate = datetime(year=int(ref[:4]), month=int(ref[4:6]), day=int(ref[6:8]), hour=hour, minute=minute)

    elif ref[:3] in months:  # MonthName DayOfMonth
        d = None
        if ref[-2:].isdigit():
            d = int(ref[-2:])
        elif ref[-1:].isdigit():
            d = int(ref[-1:])
        else:
            raise Exception("Day of month required after month name")
        refDate = datetime(year=refDate.year, month=months.index(ref[:3]) + 1, day=d, hour=hour, minute=minute)

    elif ref[:3] in weekdays:  # DayOfWeek (Monday, etc)
        todayDayName = refDate.strftime("%a").lower()[:3]
        today = weekdays.index( todayDayName )
        twoWeeks = weekdays * 2
        dayOffset = today - twoWeeks.index(ref[:3])
        if dayOffset < 0:
            dayOffset += 7
        refDate -= timedelta(days=dayOffset)

    elif ref:
        raise ValueError("Unknown day reference: %s" % rawRef)

    return tzinfo.localize(refDate)


def parseTimeOffset(offset):
    if not offset:
        return timedelta()

    t = timedelta()

    if offset[0].isdigit():
        sign = 1
    else:
        try:
            sign = { '+' : 1, '-' : -1 }[offset[0]]
        except KeyError:
            raise KeyError('Invalid offset: %s' % offset)
        offset = offset[1:]

    while offset:
        i = 1
        while offset[:i].isdigit() and i <= len(offset):
            i += 1
        num = int(offset[:i-1])
        offset = offset[i-1:]
        i = 1
        while offset[:i].isalpha() and i <= len(offset):
            i += 1
        unit = offset[:i-1]
        offset = offset[i-1:]
        unitString = getUnitString(unit)
        if unitString == MONTHS_STRING:
            unitString = DAYS_STRING
            num = num * 30
        if unitString == YEARS_STRING:
            unitString = DAYS_STRING
            num = num * 365
        t += timedelta(**{ unitString : sign * num})

    return t


def getUnitString(s):
    if s.startswith('s'): return SECONDS_STRING
    if s.startswith('min'): return MINUTES_STRING
    if s.startswith('h'): return HOURS_STRING
    if s.startswith('d'): return DAYS_STRING
    if s.startswith('w'): return WEEKS_STRING
    if s.startswith('mon'): return MONTHS_STRING
    if s.startswith('y'): return YEARS_STRING
    raise ValueError("Invalid offset unit '%s'" % s)



import sys
import os
import time

# Use the built-in version of scandir/stat if possible, otherwise
# use the scandir module version
try:
    from os import scandir, stat  # noqa # pylint: disable=unused-import
except ImportError:
    from scandir import scandir, stat  # noqa # pylint: disable=unused-import

try:
    import rrdtool
except ImportError:
    rrdtool = False

from django.conf import settings
from graphite.intervals import Interval, IntervalSet
from graphite.readers.utils import BaseReader


class RRDReader(BaseReader):
    supported = bool(rrdtool)

    @staticmethod
    def _convert_fs_path(fs_path):
        # Only Python 2 'unicode' needs to be converted to str/bytes
        try:
            if isinstance(fs_path, unicode):
                fs_path = fs_path.encode(sys.getfilesystemencoding())
        except NameError:
            pass
        return os.path.realpath(fs_path)

    def __init__(self, fs_path, datasource_name):
        self.fs_path = RRDReader._convert_fs_path(fs_path)
        self.datasource_name = datasource_name

    def get_intervals(self):
        start = time.time() - self.get_retention(self.fs_path)
        end = max(stat(self.fs_path).st_mtime, start)
        return IntervalSet([Interval(start, end)])

    def fetch(self, startTime, endTime):
        startString = time.strftime(
            "%H:%M_%Y%m%d+%Ss", time.localtime(startTime))
        endString = time.strftime("%H:%M_%Y%m%d+%Ss", time.localtime(endTime))

        if settings.FLUSHRRDCACHED:
            rrdtool.flushcached(self.fs_path, '--daemon',
                                settings.FLUSHRRDCACHED)

        (timeInfo, columns, rows) = rrdtool.fetch(
            self.fs_path,
            settings.RRD_CF, '-s' + startString, '-e' + endString)
        colIndex = list(columns).index(self.datasource_name)
        rows.pop()  # chop off the latest value because RRD returns crazy last values sometimes
        values = (row[colIndex] for row in rows)

        return (timeInfo, values)

    @staticmethod
    def get_datasources(fs_path):
        info = rrdtool.info(RRDReader._convert_fs_path(fs_path))

        if 'ds' in info:
            return [datasource_name for datasource_name in info['ds']]
        else:
            ds_keys = [key for key in info if key.startswith('ds[')]
            datasources = set(key[3:].split(']')[0] for key in ds_keys)
            return list(datasources)

    @staticmethod
    def get_retention(fs_path):
        info = rrdtool.info(RRDReader._convert_fs_path(fs_path))
        if 'rra' in info:
            rras = info['rra']
        else:
            # Ugh, I like the old python-rrdtool api better..
            rra_count = max([int(key[4])
                             for key in info if key.startswith('rra[')]) + 1
            rras = [{}] * rra_count
            for i in range(rra_count):
                rras[i]['pdp_per_row'] = info['rra[%d].pdp_per_row' % i]
                rras[i]['rows'] = info['rra[%d].rows' % i]

        retention_points = 0
        for rra in rras:
            points = rra['pdp_per_row'] * rra['rows']
            if points > retention_points:
                retention_points = points

        return retention_points * info['step']



import re
import six

from graphite.errors import NormalizeEmptyResultError, InputParameterError
from graphite.functions import SeriesFunction
from graphite.logger import log
from graphite.render.grammar import grammar
from graphite.render.datalib import fetchData, TimeSeries, prefetchData
from graphite.functions.params import validateParams

from django.conf import settings


def evaluateTarget(requestContext, targets):
  if not isinstance(targets, list):
    targets = [targets]

  pathExpressions = extractPathExpressions(requestContext, targets)
  prefetchData(requestContext, pathExpressions)

  seriesList = []

  for target in targets:
    if not target:
      continue

    if isinstance(target, six.string_types):
      if not target.strip():
        continue

      target = grammar.parseString(target)

    try:
      result = evaluateTokens(requestContext, target)
    except InputParameterError as e:
      e.setTargets(requestContext.get('targets', []))
      e.setSourceIdHeaders(requestContext.get('sourceIdHeaders', {}))
      raise

    # we have to return a list of TimeSeries objects
    if isinstance(result, TimeSeries):
      seriesList.append(result)
    elif result:
      seriesList.extend(result)

  return seriesList


def evaluateTokens(requestContext, tokens, replacements=None, pipedArg=None):
  if tokens.template:
    arglist = dict()
    if tokens.template.kwargs:
      arglist.update(dict([(kwarg.argname, evaluateScalarTokens(kwarg.args[0])) for kwarg in tokens.template.kwargs]))
    if tokens.template.args:
      arglist.update(dict([(str(i+1), evaluateScalarTokens(arg)) for i, arg in enumerate(tokens.template.args)]))
    if 'template' in requestContext:
      arglist.update(requestContext['template'])
    return evaluateTokens(requestContext, tokens.template, arglist)

  if tokens.expression:
    if tokens.expression.pipedCalls.asList():
      # when the expression has piped calls, we pop the right-most call and pass the remaining
      # expression into it via pipedArg, to get the same result as a nested call
      rightMost = tokens.expression.pipedCalls.pop()
      return evaluateTokens(requestContext, rightMost, replacements, tokens)

    return evaluateTokens(requestContext, tokens.expression, replacements)

  if tokens.pathExpression:
    expression = tokens.pathExpression
    if replacements:
      for name in replacements:
        if expression == '$'+name:
          val = replacements[name]
          if not isinstance(val, six.string_types):
            return val
          elif re.match(r'^-?[\d.]+$', val):
            return float(val)
          else:
            return val
        else:
          expression = expression.replace('$'+name, str(replacements[name]))
    return fetchData(requestContext, expression)

  if tokens.call:
    if tokens.call.funcname == 'template':
      # if template propagates down here, it means the grammar didn't match the invocation
      # as tokens.template. this generally happens if you try to pass non-numeric/string args
      raise InputParameterError("invalid template() syntax, only string/numeric arguments are allowed")

    if tokens.call.funcname == 'seriesByTag':
      return fetchData(requestContext, tokens.call.raw)

    try:
      func = SeriesFunction(tokens.call.funcname)
    except KeyError:
      raise InputParameterError('Received request for unknown function: {func}'.format(func=tokens.call.funcname))

    rawArgs = tokens.call.args or []
    if pipedArg is not None:
      rawArgs.insert(0, pipedArg)
    args = [evaluateTokens(requestContext, arg, replacements) for arg in rawArgs]
    requestContext['args'] = rawArgs
    kwargs = dict([(kwarg.argname, evaluateTokens(requestContext, kwarg.args[0], replacements))
                   for kwarg in tokens.call.kwargs])

    if hasattr(func, 'params'):
      try:
        (args, kwargs) = validateParams(tokens.call.funcname, func.params, args, kwargs)
      except InputParameterError as e:
        e.setSourceIdHeaders(requestContext.get('sourceIdHeaders', {}))
        e.setTargets(requestContext.get('targets', []))
        e.setFunction(tokens.call.funcname, args, kwargs)
        if settings.ENFORCE_INPUT_VALIDATION:
          raise
        else:
          log.warning('Validation Error: %s', str(e))

    try:
      return func(requestContext, *args, **kwargs)
    except NormalizeEmptyResultError:
      return []
    except InputParameterError as e:
        e.setSourceIdHeaders(requestContext.get('sourceIdHeaders', {}))
        e.setTargets(requestContext.get('targets', []))
        e.setFunction(tokens.call.funcname, args, kwargs)
        raise

  return evaluateScalarTokens(tokens)


def evaluateScalarTokens(tokens):
  if tokens.number:
    if tokens.number.integer:
      return int(tokens.number.integer)
    if tokens.number.float:
      return float(tokens.number.float)
    if tokens.number.scientific:
      return float(tokens.number.scientific[0])

    raise InputParameterError("unknown numeric type in target evaluator")

  if tokens.string:
    return tokens.string[1:-1]

  if tokens.boolean:
    return tokens.boolean[0] == 'true'

  if tokens.none:
    return None

  if tokens.infinity:
    return float('inf')

  raise InputParameterError("unknown token in target evaluator")


def extractPathExpressions(requestContext, targets):
  # Returns a list of unique pathExpressions found in the targets list

  pathExpressions = set()

  def extractPathExpression(requestContext, tokens, replacements=None):
    if tokens.template:
      arglist = dict()
      if tokens.template.kwargs:
        arglist.update(dict([(kwarg.argname, evaluateScalarTokens(kwarg.args[0])) for kwarg in tokens.template.kwargs]))
      if tokens.template.args:
        arglist.update(dict([(str(i+1), evaluateScalarTokens(arg)) for i, arg in enumerate(tokens.template.args)]))
      if 'template' in requestContext:
        arglist.update(requestContext['template'])
      extractPathExpression(requestContext, tokens.template, arglist)
    elif tokens.expression:
      extractPathExpression(requestContext, tokens.expression, replacements)
      if tokens.expression.pipedCalls:
        for token in tokens.expression.pipedCalls:
          extractPathExpression(requestContext, token, replacements)
    elif tokens.pathExpression:
      expression = tokens.pathExpression
      if replacements:
        for name in replacements:
          if expression != '$'+name:
            expression = expression.replace('$'+name, str(replacements[name]))
      pathExpressions.add(expression)
    elif tokens.call:
      # if we're prefetching seriesByTag, pass the entire call back as a path expression
      if tokens.call.funcname == 'seriesByTag':
        pathExpressions.add(tokens.call.raw)
      else:
        for a in tokens.call.args:
          extractPathExpression(requestContext, a, replacements)

  for target in targets:
    if not target:
      continue

    if isinstance(target, six.string_types):
      if not target.strip():
        continue
      target = grammar.parseString(target)
    extractPathExpression(requestContext, target)

  return list(pathExpressions)



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import os
from graphite.user_util import getProfile
from graphite.logger import log
from graphite.account.models import MyGraph

from django.shortcuts import render
from django.http import HttpResponse
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist


def composer(request):
    profile = getProfile(request)
    context = {
        'queryString' : request.GET.urlencode().replace('+','%20'),
        'showTarget' : request.GET.get('showTarget',''),
        'user' : request.user,
        'profile' : profile,
        'showMyGraphs' : int( profile.user.username != 'default' ),
        'searchEnabled' : int( os.access(settings.INDEX_FILE, os.R_OK) ),
        'refreshInterval': settings.AUTO_REFRESH_INTERVAL,
        'debug' : settings.DEBUG,
        'jsdebug' : settings.DEBUG,
    }
    return render(request, "composer.html", context)


def mygraph(request):
    profile = getProfile(request, allowDefault=False)

    if not profile:
        return HttpResponse("You are not logged in!")

    action = request.GET['action']
    graphName = request.GET['graphName']

    if not graphName:
        return HttpResponse("You must type in a graph name.")

    if action == 'save':
        url = request.GET['url']

        try:
            existingGraph = profile.mygraph_set.get(name=graphName)
            existingGraph.url = url
            existingGraph.save()

        except ObjectDoesNotExist:
            try:
                newGraph = MyGraph(profile=profile,name=graphName,url=url)
                newGraph.save()
            except Exception:
                log.exception("Failed to create new MyGraph in /composer/mygraph/, graphName=%s" % graphName)
                return HttpResponse("Failed to save graph %s" % graphName)

        return HttpResponse("SAVED")

    elif action == 'delete':
        try:
            existingGraph = profile.mygraph_set.get(name=graphName)
            existingGraph.delete()

        except ObjectDoesNotExist:
            return HttpResponse("No such graph '%s'" % graphName)

        return HttpResponse("DELETED")

    else:
        return HttpResponse("Invalid operation '%s'" % action)



from __future__ import absolute_import

from binascii import b2a_base64
import sys

from graphite.http_pool import http
from graphite.util import json

from graphite.tags.base import BaseTagDB


class HttpTagDB(BaseTagDB):
    """
    Stores tag information using an external http service that implements the graphite tags API.
    """
    def __init__(self, settings, *args, **kwargs):
        super(HttpTagDB, self).__init__(settings, *args, **kwargs)

        self.base_url = settings.TAGDB_HTTP_URL
        self.username = settings.TAGDB_HTTP_USER
        self.password = settings.TAGDB_HTTP_PASSWORD

    def request(self, method, url, fields, requestContext=None):
        headers = requestContext.get('forwardHeaders') if requestContext else {}
        if 'Authorization' not in headers and self.username and self.password:
            user_pw = '%s:%s' % (self.username, self.password)
            if sys.version_info[0] >= 3:
                user_pw_b64 = b2a_base64(user_pw.encode('utf-8')).decode('ascii')
            else:
                user_pw_b64 = user_pw.encode('base64')
            headers['Authorization'] = 'Basic ' + user_pw_b64

        req_fields = []
        for (field, value) in fields.items():
            if value is None:
                continue

            if isinstance(value, list) or isinstance(value, tuple):
                req_fields.extend([(field, v) for v in value if v is not None])
            else:
                req_fields.append((field, value))

        result = http.request(
            method,
            self.base_url + url,
            fields=req_fields,
            headers=headers,
            timeout=self.settings.FIND_TIMEOUT,
        )

        if result.status == 400:
            raise ValueError(json.loads(result.data.decode('utf-8')).get('error'))

        if result.status != 200:
            raise Exception('HTTP Error from remote tagdb: %s %s' % (result.status, result.data))

        return json.loads(result.data.decode('utf-8'))

    def find_series_cachekey(self, tags, requestContext=None):
        headers = [
            header + '=' + value
            for (header, value)
            in (requestContext or {}).get('forwardHeaders', {}).items()
        ]

        return 'TagDB.find_series:' + ':'.join(sorted(tags)) + ':' + ':'.join(sorted(headers))

    def _find_series(self, tags, requestContext=None):
        return self.request(
            'POST',
            '/tags/findSeries',
            {'expr': tags},
            requestContext=requestContext,
        )

    def get_series(self, path, requestContext=None):
        try:
            parsed = self.parse(path)
        except Exception:
            return None

        seriesList = self.find_series(
            [('%s=%s' % (tag, parsed.tags[tag])) for tag in parsed.tags],
            requestContext=requestContext,
        )

        if parsed.path in seriesList:
            return parsed

    def list_tags(self, tagFilter=None, limit=None, requestContext=None):
        return self.request('GET', '/tags', {'filter': tagFilter, 'limit': limit}, requestContext)

    def get_tag(self, tag, valueFilter=None, limit=None, requestContext=None):
        return self.request('GET', '/tags/' + tag, {'filter': valueFilter, 'limit': limit}, requestContext)

    def list_values(self, tag, valueFilter=None, limit=None, requestContext=None):
        tagInfo = self.get_tag(tag, valueFilter=valueFilter, limit=limit, requestContext=requestContext)
        if not tagInfo:
            return []

        return tagInfo['values']

    def tag_series(self, series, requestContext=None):
        return self.request('POST', '/tags/tagSeries', {'path': series}, requestContext)

    def tag_multi_series(self, seriesList, requestContext=None):
        return self.request('POST', '/tags/tagMultiSeries', {'path': seriesList}, requestContext)

    def del_series(self, series, requestContext=None):
        return self.request('POST', '/tags/delSeries', {'path': series}, requestContext)

    def del_multi_series(self, seriesList, requestContext=None):
        return self.request('POST', '/tags/delSeries', {'path': seriesList}, requestContext)

    def auto_complete_tags(self, exprs, tagPrefix=None, limit=None, requestContext=None):
        """
        Return auto-complete suggestions for tags based on the matches
        for the specified expressions, optionally filtered by tag prefix
        """
        if not self.settings.TAGDB_HTTP_AUTOCOMPLETE:
            return super(HttpTagDB, self).auto_complete_tags(
                exprs, tagPrefix=tagPrefix, limit=limit, requestContext=requestContext
            )

        if limit is None:
            limit = self.settings.TAGDB_AUTOCOMPLETE_LIMIT

        fields = {
            'tagPrefix': tagPrefix or '',
            'limit': str(limit),
            'expr': exprs,
        }

        return self.request('POST', '/tags/autoComplete/tags', fields, requestContext)

    def auto_complete_values(self, exprs, tag, valuePrefix=None, limit=None, requestContext=None):
        """
        Return auto-complete suggestions for tags and values based on the matches
        for the specified expressions, optionally filtered by tag and/or value prefix
        """
        if not self.settings.TAGDB_HTTP_AUTOCOMPLETE:
            return super(HttpTagDB, self).auto_complete_values(
                exprs, tag, valuePrefix=valuePrefix, limit=limit, requestContext=requestContext
            )

        if limit is None:
            limit = self.settings.TAGDB_AUTOCOMPLETE_LIMIT

        fields = {
            'tag': tag or '',
            'valuePrefix': valuePrefix or '',
            'limit': str(limit),
            'expr': exprs,
        }

        return self.request('POST', '/tags/autoComplete/values', fields, requestContext)



#!/usr/bin/env python

from __future__ import with_statement

import os
try:
    from ConfigParser import ConfigParser, DuplicateSectionError  # Python 2
except ImportError:
    from configparser import ConfigParser, DuplicateSectionError  # Python 3

from glob import glob
from collections import defaultdict

# io.StringIO is strictly unicode only. Python 2 StringIO.StringIO accepts
# bytes, so we'll conveniently ignore decoding and reencoding the file there.
try:
    from StringIO import StringIO  # Python 2
except ImportError:
    from io import StringIO  # Python 3

# Graphite historically has an install prefix set in setup.cfg. Being in a
# configuration file, it's not easy to override it or unset it (for installing
# graphite in a virtualenv for instance).
# The prefix is now set by ``setup.py`` and *unset* if an environment variable
# named ``GRAPHITE_NO_PREFIX`` is present.
# While ``setup.cfg`` doesn't contain the prefix anymore, the *unset* step is
# required for installations from a source tarball because running
# ``python setup.py sdist`` will re-add the prefix to the tarball's
# ``setup.cfg``.
with open('setup.cfg', 'r') as f:
    orig_setup_cfg = f.read()
cf = ConfigParser()
cf.readfp(StringIO(orig_setup_cfg), 'setup.cfg')

if os.environ.get('GRAPHITE_NO_PREFIX') or os.environ.get('READTHEDOCS'):
    cf.remove_section('install')
else:
    try:
        cf.add_section('install')
    except DuplicateSectionError:
        pass
    if not cf.has_option('install', 'prefix'):
        cf.set('install', 'prefix', '/opt/graphite')
    if not cf.has_option('install', 'install-lib'):
        cf.set('install', 'install-lib', '%(prefix)s/webapp')

with open('setup.cfg', 'w') as f:
    cf.write(f)

if os.environ.get('USE_SETUPTOOLS'):
    from setuptools import setup
    setup_kwargs = dict(zip_safe=0)

else:
    from distutils.core import setup
    setup_kwargs = dict()


storage_dirs = []

for subdir in ('whisper/dummy.txt', 'ceres/dummy.txt', 'rrd/dummy.txt', 'log/dummy.txt', 'log/webapp/dummy.txt'):
    storage_dirs.append( ('storage/%s' % subdir, []) )

webapp_content = defaultdict(list)

for root, dirs, files in os.walk('webapp/content'):
    for filename in files:
        filepath = os.path.join(root, filename)
        webapp_content[root].append(filepath)

conf_files = [ ('conf', glob('conf/*.example')) ]
examples = [ ('examples', glob('examples/example-*')) ]


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()


try:
    setup(
      name='graphite-web',
      version='1.1.10',
      url='http://graphiteapp.org/',
      author='Chris Davis',
      author_email='chrismd@gmail.com',
      license='Apache Software License 2.0',
      description='Enterprise scalable realtime graphing',
      long_description=read('README.md'),
      long_description_content_type='text/markdown',
      package_dir={'' : 'webapp'},
      packages=[
        'graphite',
        'graphite.account',
        'graphite.account.migrations',
        'graphite.browser',
        'graphite.composer',
        'graphite.dashboard',
        'graphite.dashboard.migrations',
        'graphite.events',
        'graphite.events.migrations',
        'graphite.finders',
        'graphite.functions',
        'graphite.functions.custom',
        'graphite.metrics',
        'graphite.readers',
        'graphite.render',
        'graphite.tags',
        'graphite.tags.migrations',
        'graphite.url_shortener',
        'graphite.url_shortener.migrations',
        'graphite.version',
        'graphite.whitelist',
        'graphite.worker_pool',
      ],
      package_data={'graphite' :
        ['templates/*', 'local_settings.py.example']},
      scripts=glob('bin/*'),
      data_files=list(webapp_content.items()) + storage_dirs + conf_files + examples,
      install_requires=['Django>=1.8,<3.1', 'django-tagging==0.4.3', 'pytz',
                        'pyparsing', 'cairocffi', 'urllib3',
                        'scandir;python_version<"3.5"', 'six'],
      classifiers=[
          'Intended Audience :: Developers',
          'Natural Language :: English',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: Implementation :: PyPy',
      ],
      **setup_kwargs
    )
finally:
    with open('setup.cfg', 'w') as f:
        f.write(orig_setup_cfg)



INFINITY = float('inf')
NEGATIVE_INFINITY = -INFINITY


class IntervalSet:

    def __init__(self, intervals, disjoint=False):
        self.intervals = intervals

        if not disjoint:
            self.intervals = union_overlapping(self.intervals)

        self.size = sum(i.size for i in self.intervals)

    def __repr__(self):
        return repr(self.intervals)

    def __iter__(self):
        return iter(self.intervals)

    def __len__(self):
        return len(self.intervals)

    def __getitem__(self, i):
        return self.intervals[i]

    def __nonzero__(self):
        return self.size != 0

    def __sub__(self, other):
        return self.intersect( other.complement() )

    def complement(self):
        complementary = []
        cursor = NEGATIVE_INFINITY

        for interval in self.intervals:
            if cursor < interval.start:
                complementary.append( Interval(cursor, interval.start) )
                cursor = interval.end

        if cursor < INFINITY:
            complementary.append( Interval(cursor, INFINITY) )

        return IntervalSet(complementary, disjoint=True)

    def intersect(self, other): #XXX The last major bottleneck. Factorial-time hell.
        # Then again, this function is entirely unused...
        if (not self) or (not other):
            return IntervalSet([])

        #earliest = max(self.intervals[0].start, other.intervals[0].start)
        #latest = min(self.intervals[-1].end, other.intervals[-1].end)

        #mine = [i for i in self.intervals if i.start >= earliest and i.end <= latest]
        #theirs = [i for i in other.intervals if i.start >= earliest and i.end <= latest]

        intersections = [x for x in (i.intersect(j)
                                     for i in self.intervals
                                     for j in other.intervals)
                         if x]

        return IntervalSet(intersections, disjoint=True)

    def intersect_interval(self, interval):
        intersections = [x for x in (i.intersect(interval)
                                     for i in self.intervals)
                         if x]
        return IntervalSet(intersections, disjoint=True)

    def union(self, other):
        return IntervalSet( sorted(self.intervals + other.intervals) )


class Interval:

    def __init__(self, start, end):
        if end - start < 0:
            raise ValueError("Invalid interval start=%s end=%s" % (start, end))

        self.start = start
        self.end = end
        self.tuple = (start, end)
        self.size = self.end - self.start

    def __eq__(self, other):
        return self.tuple == other.tuple

    def __ne__(self, other):
        return self.tuple != other.tuple

    def __hash__(self):
        return hash( self.tuple )

    def __lt__(self, other):
        return self.start < self.start

    def __le__(self, other):
        return self.start <= self.start

    def __gt__(self, other):
        return self.start > self.start

    def __ge__(self, other):
        return self.start >= self.start

    def __cmp__(self, other):
        return (self.start > other.start) - (self.start < other.start)

    def __len__(self):
        raise TypeError("len() doesn't support infinite values, use the 'size' attribute instead")

    def __nonzero__(self):  # Python 2
        return self.size != 0

    def __bool__(self):  # Python 3
        return self.size != 0

    def __repr__(self):
        return '<Interval: %s>' % str(self.tuple)

    def intersect(self, other):
        start = max(self.start, other.start)
        end = min(self.end, other.end)

        if end > start:
            return Interval(start, end)

    def overlaps(self, other):
        earlier = self if self.start <= other.start else other
        later = self if earlier is other else other
        return earlier.end >= later.start

    def union(self, other):
        if not self.overlaps(other):
            raise TypeError("Union of disjoint intervals is not an interval")

        start = min(self.start, other.start)
        end = max(self.end, other.end)
        return Interval(start, end)


def union_overlapping(intervals):
    """Union any overlapping intervals in the given set."""
    disjoint_intervals = []

    for interval in intervals:
        if disjoint_intervals and disjoint_intervals[-1].overlaps(interval):
            disjoint_intervals[-1] = disjoint_intervals[-1].union(interval)
        else:
            disjoint_intervals.append(interval)

    return disjoint_intervals



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
import csv
import math
import pytz
import six.moves.http_client

from datetime import datetime
from time import time
from random import shuffle
from six.moves.urllib.parse import urlencode, urlsplit, urlunsplit, parse_qs

from graphite.compat import HttpResponse
from graphite.errors import InputParameterError, handleInputParameterError
from graphite.user_util import getProfileByUsername
from graphite.util import json, unpickle, pickle, msgpack, BytesIO
from graphite.storage import extractForwardHeaders
from graphite.logger import log
from graphite.render.evaluator import evaluateTarget
from graphite.render.attime import parseATTime
from graphite.functions import loadFunctions, PieFunction
from graphite.render.hashing import hashRequest, hashData
from graphite.render.glyph import GraphTypes
from graphite.tags.models import Series, Tag, TagValue, SeriesTag  # noqa # pylint: disable=unused-import

from django.http import HttpResponseServerError, HttpResponseRedirect
from django.template import Context, loader
from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist
from django.conf import settings
from django.utils.cache import add_never_cache_headers, patch_response_headers
from six.moves import zip


loadFunctions()


@handleInputParameterError
def renderView(request):
  start = time()

  try:
    # we consider exceptions thrown by the option
    # parsing to be due to user input error
    (graphOptions, requestOptions) = parseOptions(request)
  except Exception as e:
    raise InputParameterError(str(e))

  useCache = 'noCache' not in requestOptions
  cacheTimeout = requestOptions['cacheTimeout']
  # TODO: Make that a namedtuple or a class.
  requestContext = {
    'startTime' : requestOptions['startTime'],
    'endTime' : requestOptions['endTime'],
    'now': requestOptions['now'],
    'localOnly' : requestOptions['localOnly'],
    'template' : requestOptions['template'],
    'tzinfo' : requestOptions['tzinfo'],
    'forwardHeaders': requestOptions['forwardHeaders'],
    'sourceIdHeaders': requestOptions['sourceIdHeaders'],
    'data' : [],
    'prefetched' : {},
    'xFilesFactor' : requestOptions['xFilesFactor'],
    'maxDataPoints' : requestOptions.get('maxDataPoints', None),
    'targets': requestOptions['targets'],
    'maxStep': requestOptions.get('maxStep', None),
  }
  data = requestContext['data']

  response = None

  # First we check the request cache
  if useCache:
    requestKey = hashRequest(request)
    response = cache.get(requestKey)
    if response:
      log.cache('Request-Cache hit [%s]' % requestKey)
      log.rendering('Returned cached response in %.6f' % (time() - start))
      return response

    log.cache('Request-Cache miss [%s]' % requestKey)

  # Now we prepare the requested data
  if requestOptions['graphType'] == 'pie':
    for target in requestOptions['targets']:
      if target.find(':') >= 0:
        try:
          name, value = target.split(':', 1)
          value = float(value)
        except ValueError:
          raise ValueError("Invalid target '%s'" % target)
        data.append((name, value))
      else:
        seriesList = evaluateTarget(requestContext, target)

        for series in seriesList:
          func = PieFunction(requestOptions['pieMode'])
          data.append( (series.name, func(requestContext, series) or 0 ))

  elif requestOptions['graphType'] == 'line':
    # Let's see if at least our data is cached
    cachedData = None
    if useCache:
      targets = requestOptions['targets']
      startTime = requestOptions['startTime']
      endTime = requestOptions['endTime']
      dataKey = hashData(targets, startTime, endTime, requestOptions['xFilesFactor'])
      cachedData = cache.get(dataKey)
      if cachedData:
        log.cache("Data-Cache hit [%s]" % dataKey)
      else:
        log.cache("Data-Cache miss [%s]" % dataKey)

    if cachedData is not None:
      requestContext['data'] = data = cachedData
    else: # Have to actually retrieve the data now
      targets = requestOptions['targets']

      data.extend(evaluateTarget(requestContext, targets))

      if useCache:
        cache.add(dataKey, data, cacheTimeout)

    renderStart = time()

    format = requestOptions.get('format')
    if format == 'csv':
      response = renderViewCsv(requestOptions, data)
    elif format == 'json':
      response = renderViewJson(requestOptions, data)
    elif format == 'dygraph':
      response = renderViewDygraph(requestOptions, data)
    elif format == 'rickshaw':
      response = renderViewRickshaw(requestOptions, data)
    elif format == 'raw':
      response = renderViewRaw(requestOptions, data)
    elif format == 'pickle':
      response = renderViewPickle(requestOptions, data)
    elif format == 'msgpack':
      response = renderViewMsgPack(requestOptions, data)

  # if response wasn't generated above, render a graph image
  if not response:
    format = 'image'
    renderStart = time()
    response = renderViewGraph(graphOptions, requestOptions, data)

  if useCache:
    cache.add(requestKey, response, cacheTimeout)
    patch_response_headers(response, cache_timeout=cacheTimeout)
  else:
    add_never_cache_headers(response)

  log.rendering('%s rendering time %6f' % (format, time() - renderStart))
  log.rendering('Total request processing time %6f' % (time() - start))

  return response


def renderViewGraph(graphOptions, requestOptions, data):
  # We've got the data, now to render it
  graphOptions['data'] = data
  if settings.REMOTE_RENDERING: # Rendering on other machines is faster in some situations
    image = delegateRendering(requestOptions['graphType'], graphOptions, requestOptions['forwardHeaders'])
  else:
    image = doImageRender(requestOptions['graphClass'], graphOptions)

  if graphOptions['outputFormat'] == 'pdf':
    return buildResponse(image, 'application/x-pdf')

  if graphOptions['outputFormat'] == 'svg':
    if 'jsonp' in requestOptions:
      return HttpResponse(
        content="%s(%s)" % (requestOptions['jsonp'], json.dumps(image)),
        content_type='text/javascript')

    return buildResponse(image, 'image/svg+xml')

  return buildResponse(image, 'image/png')


def renderViewCsv(requestOptions, data):
  response = HttpResponse(content_type='text/csv')
  writer = csv.writer(response, dialect='excel')

  for series in data:
    for i, value in enumerate(series):
      timestamp = datetime.fromtimestamp(series.start + (i * series.step), requestOptions['tzinfo'])
      writer.writerow((series.name, timestamp.strftime("%Y-%m-%d %H:%M:%S"), value))

  return response


def renderViewJson(requestOptions, data):
  series_data = []

  if any(data):
    startTime = min([series.start for series in data])
    endTime = max([series.end for series in data])
    timeRange = endTime - startTime

    for series in data:
      if 'maxDataPoints' in requestOptions:
        maxDataPoints = requestOptions['maxDataPoints']
        if maxDataPoints == 1:
          series.consolidate(len(series))
        else:
          numberOfDataPoints = timeRange/series.step
          if maxDataPoints < numberOfDataPoints:
            valuesPerPoint = math.ceil(float(numberOfDataPoints) / float(maxDataPoints))
            secondsPerPoint = int(valuesPerPoint * series.step)
            # Nudge start over a little bit so that the consolidation bands align with each call
            # removing 'jitter' seen when refreshing.
            nudge = secondsPerPoint + (series.start % series.step) - (series.start % secondsPerPoint)
            series.start = series.start + nudge
            valuesToLose = int(nudge/series.step)
            for r in range(1, valuesToLose):
              del series[0]
            series.consolidate(valuesPerPoint)

      datapoints = series.datapoints()

      if 'noNullPoints' in requestOptions:
        datapoints = [
          point for point in datapoints
          if point[0] is not None and not math.isnan(point[0])
        ]
        if not datapoints:
          continue

      series_data.append(dict(target=series.name, tags=dict(series.tags), datapoints=datapoints))

  output = json.dumps(series_data, indent=(2 if requestOptions.get('pretty') else None)) \
      .replace('NaN,', 'null,').replace('Infinity,', '1e9999,')

  if 'jsonp' in requestOptions:
    response = HttpResponse(
      content="%s(%s)" % (requestOptions['jsonp'], output),
      content_type='text/javascript')
  else:
    response = HttpResponse(
      content=output,
      content_type='application/json')

  return response


def renderViewDygraph(requestOptions, data):
  labels = ['Time']
  output = '{}'
  if data:
    datapoints = [[ts] for ts in range(data[0].start, data[0].end, data[0].step)]
    for series in data:
      labels.append(series.name)
      for i, point in enumerate(series):
        if point is None:
          point = 'null'
        elif point == float('inf'):
          point = 'Infinity'
        elif point == float('-inf'):
          point = '-Infinity'
        elif math.isnan(point):
          point = 'null'
        datapoints[i].append(point)
    line_template = '[%%s000%s]' % ''.join([', %s'] * len(data))
    lines = [line_template % tuple(points) for points in datapoints]
    output = '{"labels" : %s, "data" : [%s]}' % (json.dumps(labels), ', '.join(lines))

  if 'jsonp' in requestOptions:
    response = HttpResponse(
      content="%s(%s)" % (requestOptions['jsonp'], output),
      content_type='text/javascript')
  else:
    response = HttpResponse(
      content=output,
      content_type='application/json')

  return response


def renderViewRickshaw(requestOptions, data):
  series_data = []
  for series in data:
    timestamps = range(series.start, series.end, series.step)
    datapoints = [{'x' : x, 'y' : y} for x, y in zip(timestamps, series)]
    series_data.append( dict(target=series.name, datapoints=datapoints) )

  output = json.dumps(series_data, indent=(2 if requestOptions.get('pretty') else None))

  if 'jsonp' in requestOptions:
    response = HttpResponse(
      content="%s(%s)" % (requestOptions['jsonp'], output),
      content_type='text/javascript')
  else:
    response = HttpResponse(
      content=output,
      content_type='application/json')

  return response


def renderViewRaw(requestOptions, data):
  response = HttpResponse(content_type='text/plain')

  for series in data:
    response.write( "%s,%d,%d,%d|" % (series.name, series.start, series.end, series.step) )
    response.write( ','.join(map(repr,series)) )
    response.write('\n')

  return response


def renderViewPickle(requestOptions, data):
  response = HttpResponse(content_type='application/pickle')
  seriesInfo = [series.getInfo() for series in data]
  pickle.dump(seriesInfo, response, protocol=-1)
  return response


def renderViewMsgPack(requestOptions, data):
  response = HttpResponse(content_type='application/x-msgpack')
  seriesInfo = [series.getInfo() for series in data]
  msgpack.dump(seriesInfo, response, use_bin_type=True)
  return response


def parseOptions(request):
  queryParams = request.GET.copy()
  queryParams.update(request.POST)

  # Start with some defaults
  graphOptions = {'width' : 330, 'height' : 250}
  requestOptions = {}

  graphType = queryParams.get('graphType','line')
  if graphType not in GraphTypes:
    raise AssertionError("Invalid graphType '%s', must be one of %s"
                         % (graphType,list(GraphTypes)))
  graphClass = GraphTypes[graphType]

  # Fill in the requestOptions
  requestOptions['graphType'] = graphType
  requestOptions['graphClass'] = graphClass
  requestOptions['pieMode'] = queryParams.get('pieMode', 'average')
  cacheTimeout = int( queryParams.get('cacheTimeout', settings.DEFAULT_CACHE_DURATION) )
  requestOptions['targets'] = []
  requestOptions['forwardHeaders'] = extractForwardHeaders(request)
  requestOptions['sourceIdHeaders'] = extractSourceIdHeaders(request)

  # Extract the targets out of the queryParams
  mytargets = []
  # Normal format: ?target=path.1&target=path.2
  if len(queryParams.getlist('target')) > 0:
    mytargets = queryParams.getlist('target')

  # Rails/PHP/jQuery common practice format: ?target[]=path.1&target[]=path.2
  elif len(queryParams.getlist('target[]')) > 0:
    mytargets = queryParams.getlist('target[]')

  # Collect the targets
  for target in mytargets:
    requestOptions['targets'].append(target)

  template = dict()
  for key, val in queryParams.items():
    if key.startswith("template["):
      template[key[9:-1]] = val
  requestOptions['template'] = template

  if 'pickle' in queryParams:
    requestOptions['format'] = 'pickle'
  if 'rawData' in queryParams:
    requestOptions['format'] = 'raw'
  if 'format' in queryParams:
    requestOptions['format'] = queryParams['format']
    if 'jsonp' in queryParams:
      requestOptions['jsonp'] = queryParams['jsonp']

  requestOptions['pretty'] = bool(queryParams.get('pretty'))

  if 'noCache' in queryParams:
    requestOptions['noCache'] = True
  if 'maxDataPoints' in queryParams and queryParams['maxDataPoints'].isdigit():
    requestOptions['maxDataPoints'] = int(queryParams['maxDataPoints'])
  if 'noNullPoints' in queryParams:
    requestOptions['noNullPoints'] = True
  if 'maxStep' in queryParams and queryParams['maxStep'].isdigit():
    requestOptions['maxStep'] = int(queryParams['maxStep'])

  requestOptions['localOnly'] = queryParams.get('local') == '1'

  # Fill in the graphOptions
  format = requestOptions.get('format')
  if format == 'svg':
    graphOptions['outputFormat'] = 'svg'
  elif format == 'pdf':
    graphOptions['outputFormat'] = 'pdf'
  else:
    graphOptions['outputFormat'] = 'png'

  for opt in graphClass.customizable:
    if opt in queryParams:
      val = queryParams[opt]
      if (val.isdigit() or (val.startswith('-') and val[1:].isdigit())) and 'color' not in opt.lower():
        val = int(val)
      elif '.' in val and (val.replace('.','',1).isdigit() or (val.startswith('-') and val[1:].replace('.','',1).isdigit())):
        val = float(val)
      elif val.lower() in ('true','false'):
        val = val.lower() == 'true'
      elif val.lower() == 'default' or val == '':
        continue
      graphOptions[opt] = val

  tzinfo = pytz.timezone(settings.TIME_ZONE)
  if 'tz' in queryParams:
    try:
      tzinfo = pytz.timezone(queryParams['tz'])
    except pytz.UnknownTimeZoneError:
      pass
  requestOptions['tzinfo'] = tzinfo

  # Get the time interval for time-oriented graph types
  if graphType == 'line' or graphType == 'pie':
    if 'now' in queryParams:
        now = parseATTime(queryParams['now'], tzinfo)
    else:
        now = datetime.now(tzinfo)

    if 'until' in queryParams:
      untilTime = parseATTime(queryParams['until'], tzinfo, now)
    else:
      untilTime = now
    if 'from' in queryParams:
      fromTime = parseATTime(queryParams['from'], tzinfo, now)
    else:
      fromTime = parseATTime('-1d', tzinfo, now)

    startTime = min(fromTime, untilTime)
    endTime = max(fromTime, untilTime)
    assert startTime != endTime, "Invalid empty time range"

    requestOptions['startTime'] = startTime
    requestOptions['endTime'] = endTime
    timeRange = endTime - startTime
    queryTime = timeRange.days * 86400 + timeRange.seconds # convert the time delta to seconds
    if settings.DEFAULT_CACHE_POLICY and not queryParams.get('cacheTimeout'):
      timeouts = [timeout for period,timeout in settings.DEFAULT_CACHE_POLICY if period <= queryTime]
      cacheTimeout = max(timeouts or (0,))
    requestOptions['now'] = now

  if cacheTimeout == 0:
    requestOptions['noCache'] = True
  requestOptions['cacheTimeout'] = cacheTimeout

  requestOptions['xFilesFactor'] = float( queryParams.get('xFilesFactor', settings.DEFAULT_XFILES_FACTOR) )

  return (graphOptions, requestOptions)


# extract headers which get set by Grafana when issuing queries, to help identifying where a query came from.
# user-defined headers from settings.INPUT_VALIDATION_SOURCE_ID_HEADERS also get extracted and mixed
# with the standard Grafana headers.
def extractSourceIdHeaders(request):
    source_headers = {
        'X-Grafana-Org-ID': 'grafana-org-id',
        'X-Dashboard-ID': 'dashboard-id',
        'X-Panel-ID': 'panel-id',
    }
    source_headers.update(settings.INPUT_VALIDATION_SOURCE_ID_HEADERS)

    headers = {}
    for hdr_name, log_name in source_headers.items():
        value = request.META.get('HTTP_' + hdr_name.upper().replace('-', '_'))
        if value:
            headers[log_name] = value

    return headers


connectionPools = {}


def connector_class_selector(https_support=False):
    return six.moves.http_client.HTTPSConnection if https_support else six.moves.http_client.HTTPConnection


def delegateRendering(graphType, graphOptions, headers=None):
  if headers is None:
    headers = {}
  start = time()
  postData = (graphType + '\n').encode() + pickle.dumps(graphOptions)
  servers = settings.RENDERING_HOSTS[:] #make a copy so we can shuffle it safely
  shuffle(servers)
  connector_class = connector_class_selector(settings.INTRACLUSTER_HTTPS)
  for server in servers:
    start2 = time()
    try:
      # Get a connection
      try:
        pool = connectionPools[server]
      except KeyError: #happens the first time
        pool = connectionPools[server] = set()
      try:
        connection = pool.pop()
      except KeyError: #No available connections, have to make a new one
        connection = connector_class(server)
        connection.timeout = settings.REMOTE_RENDER_CONNECT_TIMEOUT
      # Send the request
      try:
        connection.request('POST','/render/local/', postData, headers)
      except six.moves.http_client.CannotSendRequest:
        connection = connector_class(server) #retry once
        connection.timeout = settings.REMOTE_RENDER_CONNECT_TIMEOUT
        connection.request('POST', '/render/local/', postData, headers)
      # Read the response
      try: # Python 2.7+, use buffering of HTTP responses
        response = connection.getresponse(buffering=True)
      except TypeError:  # Python 2.6 and older
        response = connection.getresponse()
      assert response.status == 200, "Bad response code %d from %s" % (response.status,server)
      contentType = response.getheader('Content-Type')
      imageData = response.read()
      assert contentType == 'image/png', "Bad content type: \"%s\" from %s" % (contentType,server)
      assert imageData, "Received empty response from %s" % server
      # Wrap things up
      log.rendering('Remotely rendered image on %s in %.6f seconds' % (server,time() - start2))
      log.rendering('Spent a total of %.6f seconds doing remote rendering work' % (time() - start))
      pool.add(connection)
      return imageData
    except Exception:
      log.exception("Exception while attempting remote rendering request on %s" % server)
      log.rendering('Exception while remotely rendering on %s wasted %.6f' % (server,time() - start2))
      continue


def renderLocalView(request):
  try:
    start = time()
    reqParams = BytesIO(request.body)
    graphType = reqParams.readline().strip().decode()
    optionsPickle = reqParams.read()
    reqParams.close()
    graphClass = GraphTypes[graphType]
    options = unpickle.loads(optionsPickle)
    image = doImageRender(graphClass, options)
    log.rendering("Delegated rendering request took %.6f seconds" % (time() -  start))
    response = buildResponse(image)
    add_never_cache_headers(response)
    return response
  except Exception:
    log.exception("Exception in graphite.render.views.rawrender")
    return HttpResponseServerError()


def renderMyGraphView(request,username,graphName):
  profile = getProfileByUsername(username)
  if not profile:
    return errorPage("No such user '%s'" % username)
  try:
    graph = profile.mygraph_set.get(name=graphName)
  except ObjectDoesNotExist:
    return errorPage("User %s doesn't have a MyGraph named '%s'" % (username,graphName))

  request_params = request.GET.copy()
  request_params.update(request.POST)
  if request_params:
    url_parts = urlsplit(graph.url)
    query_string = url_parts[3]
    if query_string:
      url_params = parse_qs(query_string)
      # Remove lists so that we can do an update() on the dict
      for param, value in url_params.items():
        if isinstance(value, list) and param != 'target':
          url_params[param] = value[-1]
      url_params.update(request_params)
      # Handle 'target' being a list - we want duplicate &target params out of it
      url_param_pairs = []
      for key,val in url_params.items():
        if isinstance(val, list):
          for v in val:
            url_param_pairs.append( (key,v) )
        else:
          url_param_pairs.append( (key,val) )

      query_string = urlencode(url_param_pairs)
    url = urlunsplit(url_parts[:3] + (query_string,) + url_parts[4:])
  else:
    url = graph.url
  return HttpResponseRedirect(url)


def doImageRender(graphClass, graphOptions):
  pngData = BytesIO()
  t = time()
  img = graphClass(**graphOptions)
  img.output(pngData)
  log.rendering('Rendered PNG in %.6f seconds' % (time() - t))
  imageData = pngData.getvalue()
  pngData.close()
  return imageData


def buildResponse(imageData, content_type="image/png"):
  return HttpResponse(imageData, content_type=content_type)


def errorPage(message):
  template = loader.get_template('500.html')
  context = Context(dict(message=message))
  return HttpResponseServerError( template.render(context) )
